# Load utility functions from the "utils.R" file
source("utils.R")


# Define a list of simulations to run
simulations <- c('S1', 'S2', 'S3')
# Set the number of iterations and the contamination rate
N = 50
pcont = 0


################################################################################
# Code for each simulation scenario
################################################################################


# Function to generate synthetic data for Simulation 1
generate_S1 = function(pcont=0){
  n <-2000                 
  p1 <- 0.25; p2 <- 0.25; p3 <- 0.25; p4 <- 0.25                            
  mu1 <- c(0,0,0); mu2 <- c(0,2,3); mu3 <- c(3,0,-1); mu4 <- c(-3,-1,0)          
  p <-  c(p1, p2, p3,p4)
  mu <-  c(mu1, mu2, mu3,mu4)
  sigma <-  1
  
  nb = length(p)    
  y <- sample(1:nb,size=n,prob=p,replace=TRUE)       
  vec <-  sapply(1:nb,function(i){ return (sum(y==i))})    
  Sigma <- diag(c(sigma,sigma,sigma)) 
  
  X <- rmvnorm(vec[1], mu1, Sigma)      
  X <- rbind(X, rmvnorm(vec[2], mu2, Sigma))  
  X <- rbind(X, rmvnorm(vec[3], mu3, Sigma)) 
  X <- rbind(X, rmvnorm(vec[4], mu4, Sigma)) 
  
  d = ncol(X)
  if (pcont>0)
  {
    Z=matrix(rt(pcont*n*d,df=1),ncol=d)
    I=sample(1:n,size=pcont*n)
    X[I,]=Z
  }
  mel=sample.int(n)
  X=X[mel,]
  return (X)
}

# Function to generate synthetic data for Simulation 2
generate_S2 = function(pcont=0){
  n <- 2500                 
  p1 <- 0.2; p2 <- 0.2; p3 <- 0.2; p4 <- 0.2; p5 <- 0.2                            
  mu1 <- c(0,0,0,0); mu2 <- c(3,5,-1,0); mu3 <- c(-5,0,0,0); mu4 <- c(1,1,6,-2); mu5 <- c(1,-3,-2,5)          
  p <-  c(p1, p2, p3,p4,p5)
  mu <-  c(mu1, mu2, mu3,mu4,mu5)
  sigma <-  1
  
  nb = length(p)    
  y <- sample(1:nb,size=n,prob=p,replace=TRUE)       
  vec <-  sapply(1:nb,function(i){ return (sum(y==i))})    
  Sigma <- diag(c(sigma,sigma,sigma,sigma))    
  
  X <-  rmvnorm(vec[1],mu1,Sigma)      
  X <-  rbind(X ,rmvnorm(vec[2],mu2,Sigma))  
  X <-  rbind(X ,rmvnorm(vec[3],mu3,Sigma)) 
  X <-  rbind(X ,rmvnorm(vec[4],mu4,Sigma)) 
  X <-  rbind(X ,rmvnorm(vec[5],mu5,Sigma)) 
  
  d = ncol(X)
  if (pcont>0)
  {
    Z=matrix(rt(pcont*n*d,df=1),ncol=d)
    I=sample(1:n,size=pcont*n)
    X[I,]=Z
  }
  mel=sample.int(n)
  X=X[mel,]
  return (X)
}

# Function to generate synthetic data for Simulation 3
generate_S3 <- function(pcont=0){
  n <- 1500
  p1 <- 1/3; p2 <- 1/3; p3 <- 1/3
  mu1 <- c(0, 0); mu2 <- c(0, 5); mu3 <- c(5, 3)
  df <- 2
  p <- c(p1, p2, p3)
  mu <- c(mu1, mu2, mu3)
  
  nb <- length(p)
  y <- sample(1:nb, size = n, prob = p, replace = TRUE)
  vec <- sapply(1:nb, function(i) sum(y == i))
  
  Sigma <- diag(rep(1, length(mu1)))
  
  X <-  rmvt(n = vec[1], mu = mu1, sigma = Sigma, df = df)      
  X <-  rbind(X ,rmvt(n = vec[2], mu = mu2, sigma = Sigma, df = df))  
  X <-  rbind(X ,rmvt(n = vec[3], mu = mu3, sigma = Sigma, df = df)) 
  
  d = ncol(X)
  if (pcont>0)
  {
    Z=matrix(rt(pcont*n*d,df=1),ncol=d)
    I=sample(1:n,size=pcont*n)
    X[I,]=Z
  }
  mel=sample.int(n)
  X=X[mel,]
  return (X)
}


################################################################################
# Run code for each simulation and each of the methods
################################################################################


# Initialize a list to store results for each simulation
results_list <- list()

# Loop through each simulation
for (sim in simulations) {
  
  results_df <- data.frame(
    Slope_Offline = numeric(),
    Slope_Semi_Online = numeric(),
    Slope_Online = numeric(),
    Slope_Kmeans = numeric(),
    Gap_Offline = numeric(),
    Gap_Semi_Online = numeric(),
    Gap_Online = numeric(),
    Gap_Kmeans = numeric(),
    Silhouette_Offline = numeric(),
    Silhouette_Semi_Online = numeric(),
    Silhouette_Online = numeric(),
    Silhouette_Kmeans = numeric()
  )
  
  # Loop through a specified number of iterations
  for(i in 1:N){
    tryCatch({
      
      # Generate synthetic data based on the current simulation
      X <- switch(sim,
                  'S1' = generate_S1(pcont=pcont),
                  'S2' = generate_S2(pcont=pcont),
                  'S3' = generate_S3(pcont=pcont))
      
      print(i)
      
      # Perform our Slope method
      a = Kmedians(X, nclust = 1:12, method = 'Offline')
      a2 = Kmedians(X, nclust = 1:12, method = 'Semi-Online')
      a3 = Kmedians(X, nclust = 1:12, method = 'Online')
      b = Kmeans(X, nclust = 1:12, ninit = 10)
      
      # Perform Gap Statistic method
      gap <- clusGap(X, f_offline, K.max = 12, B = 100, d.power = 1)
      k.gap.offline <- maxSE(gap$Tab[, "gap"], gap$Tab[, "SE.sim"], method = "Tibs2001SEmax")
      
      gap <- clusGap(X, f_semionline, K.max = 12, B = 100, d.power = 1)
      k.gap.semi_online <- maxSE(gap$Tab[, "gap"], gap$Tab[, "SE.sim"], method = "Tibs2001SEmax")
      
      gap <- clusGap(X, f_online, K.max = 12, B = 100, d.power = 1)
      k.gap.online <-  maxSE(gap$Tab[, "gap"], gap$Tab[, "SE.sim"], method = "Tibs2001SEmax")
      
      gap <- clusGap(X, f_means, K.max = 12, B = 100, d.power = 2)
      k.gap.means <- maxSE(gap$Tab[, "gap"], gap$Tab[, "SE.sim"], method = "Tibs2001SEmax")
      
      # Perform Silhouette method
      k.sil.offline <- silhouette_offline(X,12)
      k.sil.semi_online <- silhouette_semionline(X,12)
      k.sil.online <- silhouette_online(X,12)
      k.sil.means <- silhouette_means(X,12)
      
      # Store results for the current iteration
      results_df <- rbind(results_df, data.frame(
        Slope_Offline = a$Ksel,
        Slope_Semi_Online = a2$Ksel,
        Slope_Online = a3$Ksel,
        Slope_Kmeans = b$Ksel,
        Gap_Offline = k.gap.offline,
        Gap_Semi_Online = k.gap.semi_online,
        Gap_Online = k.gap.online,
        Gap_Kmeans = k.gap.means,
        Silhouette_Offline = k.sil.offline,
        Silhouette_Semi_Online = k.sil.semi_online,
        Silhouette_Online = k.sil.online,
        Silhouette_Kmeans = k.sil.means
      ))
    }, error = function(e) {
      # Handle errors during iterations
      cat("Error in iteration ", i, ": ", conditionMessage(e), "\n")
      print(traceback())
    })
  }
  
  # Store results for the current simulation in the list
  results_list[[sim]] <- results_df
}

################################################################################
# Analysis and summary for each simulation
################################################################################

# Simulation 1
results_S1 <- results_list[['S1']]
nrow_S1 <- nrow(results_S1)
colMeans_S1 <- colMeans(results_S1)
colSums_S1 <- colSums(results_S1 == 4)
print("Simulation 1")
print(colMeans_S1)
print(colSums_S1)

# Simulation 2
results_S2 <- results_list[['S2']]
nrow_S2 <- nrow(results_S2)
colMeans_S2 <- colMeans(results_S2)
colSums_S2 <- colSums(results_S2 == 5)
print("Simulation 2")
print(colMeans_S2)
print(colSums_S2)

# Simulation 3
results_S3 <- results_list[['S3']]
nrow_S3 <- nrow(results_S3)
colMeans_S3 <- colMeans(results_S3)
colSums_S3 <- colSums(results_S3 == 3)
print("Simulation 3")
print(colMeans_S3)
print(colSums_S3)







################################################################################


sim = 'S2'
N = 20
pcont = 0


results_df <- data.frame(
  Gap_Offline = numeric(),
  Gap_Semi_Online = numeric(),
  Gap_Online = numeric(),
  Gap_Kmeans = numeric()
)

for(i in 1:N){
  tryCatch({
    
    X <- switch(sim,
                'S1' = generate_S1(pcont=pcont),
                'S2' = generate_S2(pcont=pcont),
                'S3' = generate_S3(pcont=pcont))
    
    
    print(i)
    
    # GAP STATISTIC
    gap <- clusGap(X, f_offline, K.max = 12, B = 100, d.power = 1)
    k.gap.offline <- maxSE(gap$Tab[, "gap"], gap$Tab[, "SE.sim"], method = "Tibs2001SEmax")
    
    gap <- clusGap(X, f_semionline, K.max = 12, B = 100, d.power = 1)
    k.gap.semi_online <- maxSE(gap$Tab[, "gap"], gap$Tab[, "SE.sim"], method = "Tibs2001SEmax")
    
    gap <- clusGap(X, f_online, K.max = 12, B = 100, d.power = 1)
    k.gap.online <-  maxSE(gap$Tab[, "gap"], gap$Tab[, "SE.sim"], method = "Tibs2001SEmax")
    
    gap <- clusGap(X, f_means, K.max = 12, B = 100, d.power = 2)
    k.gap.means <- maxSE(gap$Tab[, "gap"], gap$Tab[, "SE.sim"], method = "Tibs2001SEmax")
    
    
    
    results_df <- rbind(results_df, data.frame(
      Gap_Offline = k.gap.offline,
      Gap_Semi_Online = k.gap.semi_online,
      Gap_Online = k.gap.online,
      Gap_Kmeans = k.gap.means
    ))
  }, error = function(e) {
    cat("Error in iteration ", i, ": ", conditionMessage(e), "\n")
    print(traceback())
  })
}

colMeans(results_df)
colSums(results_df == 3)
