# Install and load packages
packages_to_install <- c("Gmedian", "genieclust", "doParallel", "foreach", "mclust", "factoextra",
                         "mvtnorm", "cluster", "capushe", "reshape2", "sigclust", "sigclust", "MASS", "mvnfast", "Kmedians")

#install.packages(packages_to_install)

library(Gmedian)
library(genieclust)
library(doParallel)
library(foreach)
library(genieclust)
library(mclust)
library(factoextra)
library(mvtnorm)
library(cluster)
library(capushe)
library(reshape2)
library(sigclust)
library(MASS)
library(mvnfast)
library(Kmedians)


gen_K_new=function(n=500,d=5,K=3,pcont=0,df=1,cont="Student",min=-5,max=5,radius=5)
{
  Sigma= (diag(d)*2)
  X=c()
  Tclassif=c()
  mean=rep(0,d)
  cen = c()
  for (k in 1:K)
  {
    Z=rnorm(d)
    Z=radius*Z/sqrt(sum(Z^2))
    X=rbind(X,mvtnorm::rmvnorm(n ,mean=Z))
    Tclassif=c(Tclassif,rep(k,n))
    cen = rbind(cen,Z)
  }
  
  if (pcont>0)
  {
    if(cont=='Unif')
    {
      Z=matrix(runif(pcont*n*K*d,min,max),ncol=d)
    }
    if (cont=='Student')
    {
      Z=matrix(rt(pcont*n*K*d,df=df),ncol=d)
    }
    I=sample(1:(K*n),size=pcont*n*K)
    X[I,]=Z
    Tclassif[I]="outliers"
  }
  
  
  mel=sample.int(K*n)
  Tclassif=Tclassif[mel]
  X=X[mel,]
  I = which(Tclassif != "outliers")
  return(list(X=X,classif=Tclassif,I=I,cen=cen))
}


# Gap Statistic

f_offline = function(X,k){
  l = list(cluster = KmedianK(X,k,ninit=0,init=T)$cluster)
  return (l)
}
f_semionline = function(X,k){
  l = list(cluster = KGmedianK(X,k,ninit=0,init=T)$cluster)
  return (l)
}
f_online = function(X,k){
  l = list(cluster = OnlineKmedianK(X,k,ninit=0,init=T)$cluster)
  return (l)
}
f_means = function(X,k){
  cluster=genieclust::genie(X,k=k)
  centroids = aggregate(X,list(cluster),mean)[,2:(ncol(X)+1)]
  resultatk=kmeans(X,centroids,iter.max = 30)
  l = list(cluster = resultatk$cluster)
  return (l)
}


# Silhouette

silhouette_offline = function(X,K){
  s = c()
  for (k in 2:K){
    res = KmedianK(X,k,ninit=0,init=T)
    di = dist(X, method = "euclidean", diag = TRUE, upper = TRUE, p = 1)
    sil = silhouette (res$cluster, di)
    a = mean(sil[,3])
    s = c(s,a)
  }
  return (which.max(s) + 1)
}

silhouette_semionline = function(X,K){
  s = c()
  for (k in 2:K){
    res = KGmedianK(X,k,ninit=0,init=T)
    di = dist(X, method = "euclidean", diag = TRUE, upper = TRUE, p = 1)
    sil = silhouette (res$cluster, di)
    a = mean(sil[,3])
    s = c(s,a)
  }
  return (which.max(s) + 1)
}

silhouette_online = function(X,K){
  s = c()
  for (k in 2:K){
    res = OnlineKmedianK(X,k,ninit=0,init=T)
    di = dist(X, method = "euclidean", diag = TRUE, upper = TRUE, p = 1)
    sil = silhouette (res$cluster, di)
    a = mean(sil[,3])
    s = c(s,a)
  }
  return (which.max(s) + 1)
}

silhouette_means = function(X,K){
  s = c()
  for (k in 2:K){
    cluster=genieclust::genie(X,k=k)
    centroids = aggregate(X,list(cluster),mean)[,2:(ncol(X)+1)]
    res.means = kmeans(X,centroids,iter.max = 30)
    di = dist(X, method = "euclidean", diag = TRUE, upper = TRUE, p = 2)
    sil.means = silhouette (res.means$cluster , di)
    a = mean(sil.means[,3])
    s = c(s,a)
  }
  return (which.max(s) + 1)
}


# SigClust

sigclust_k = function(X, kmax=40, alpha = 0.1){
  for (k in 2:kmax){
    a=Kmedians(X,nclust=k)
    for (l in 1:(k-1)){
      for (j in l:k){
        I = which(a$allresult$cluster==l)
        J = which(a$allresult$cluster==j)
        Xtest=X[c(I,J),]
        sig=sigclust::sigclust(Xtest,nsim=20)
        if (sig@pvalnorm > alpha)   break;
      }
      if (sig@pvalnorm > alpha)   break;
    }
    if (sig@pvalnorm > alpha)   break;
  }
  kopt=k
  return (kopt)
}








